#! /bin/bash

AVRDUDE=avrdude
PROGRAMMER=avrisp2
PORT=usb
PART=atmega644

function avrdude {
  OUTPUT=`$AVRDUDE $1`
  # Return any errors to screen
  EXIT=$?
  if [[ $EXIT != "0" ]]; then
    echo -e "$OUTPUT"
    exit $EXIT
  fi
}

while [ $# -gt 0 ]; do
  case "$1" in
    -p|--part)
      PART=$2
      shift
      ;;
    at*)
      PART=$1
      ;;
    -P|--port)
      PORT=$2
      shift
      ;;
    -c|--programmer)
      PROGRAMMER=$2
      shift
      ;;
    -r|--read-fuses)
      GETFUSES=1
      if [[ $WRITEFUSES ]]; then
        echo "You can not both read and write fuses in a single call"
        exit 1
      fi
      ;;
    -l|--lfuse)
      WRITEFUSES=1
      LFUSE=$2
      shift
      if [[ $GETFUSES ]]; then
        echo "You can not both write and read fuses in a single call"
        exit 2
      fi
      if [[ ! ${#LFUSE} -eq 4 ]]; then
        echo "Not a valid hex value for low fuse"
        exit 2
      fi
      ;;
    -h|--hfuse)
      WRITEFUSES=1
      HFUSE=$2
      shift
      if [[ $GETFUSES ]]; then
        echo "You can not both write and read fuses in a single call"
        exit 2
      fi
      if [[ ! ${#HFUSE} -eq 4 ]]; then
        echo "Not a valid hex value for high fuse"
        exit 2
      fi
      ;;
    -e|--efuse)
      WRITEFUSES=1
      EFUSE=$2
      shift
      if [[ $GETFUSES ]]; then
        echo "You can not both write and read fuses in a single call"
        exit 2
      fi
      if [[ ! ${#EFUSE} -eq 4 ]]; then
        echo "Not a valid hex value for extended fuse"
        exit 2
      fi
      ;;
    -i|--ihex)
      HEX=$2
      shift
      if [[ ! -f $HEX ]]; then
        echo "The file $HEX does not exist, maybe you still need to compile?"
        exit 3
      fi
      ;;
    *.hex)
      HEX=$1
      if [[ ! -f $HEX ]]; then
        echo "The file $HEX does not exist, maybe you still need to compile?"
        exit 3
      fi
      ;;
    *)
      echo "Not a valid option: $1"
      exit 4
      ;;
  esac
  shift
done

case $PART in
  "")
    echo "You must specify an AVR device"
    exit 5
    ;;
  "atmega128")
    NAME=ATmega128
    MCU=m128
    ;;
  "atmega1281")
    NAME=ATmega1281
    MCU=m1281
    ;;
  "attiny85")
    NAME=ATtiny85
    MCU=t85
    ;;
  "atmega16")
    NAME=ATmega16
    MCU=m16
    ;;
  "atmega644")
    NAME=ATmega644
    MCU=m644
    ;;
  "atmega644p")
    NAME=ATmega644p
    MCU=m644p
    ;;
  *)
    echo "Not a recognized device: $1"
    exit 1
    ;;
esac

CMD="-c $PROGRAMMER -p $MCU -P $PORT"

if [[ $GETFUSES ]]; then
  echo "Starting to get fuses..."
  echo "$CMD -U lfuse:r:low:h -U hfuse:r:high:h -U efuse:r:ext:h"
  avrdude "$CMD -U lfuse:r:low:h -U hfuse:r:high:h -U efuse:r:ext:h"
  echo -e "Low\tHigh\tExtended"
  read LOW < low
  read HIGH < high
  read EXT < ext
  echo -e "$LOW\t$HIGH\t$EXT"
  rm low high ext
elif [[ $WRITEFUSES ]]; then
  if [[ $LFUSE ]]; then
    CMD+=" -U lfuse:w:$LFUSE:m"
  fi
  if [[ $HFUSE ]]; then
    CMD+=" -U hfuse:w:$HFUSE:m"
  fi
  if [[ $EFUSE ]]; then
    CMD+=" -U efuse:w:$EFUSE:m"
  fi
  echo "avrdude $CMD"
  avrdude "$CMD"
elif [[ $HEX ]]; then
  echo "Programming $HEX..."
  avrdude "-c $PROGRAMMER -p $MCU -P $PORT -U flash:w:$HEX:i"
else
  echo "No action requested, doing nothing"
fi
